/*
 * LiPo Tester - Li-Ion capacity tester
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 316 2016-03-04 10:40:29Z schwenke $
 */

#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <stdio.h>

#include "lcd.h"


/* defines and macros */

#define BAUD 9600        // USART baud rate
#define LOW_VOLTAGE 3.0  // stop discharging here
#define MAGIC !0xBAD     // marker in EEPROM for valid calibration
#define OMEGA 0xF4       // the Ohm char in HD44780


/* function prototypes */

void    hw_init(void);
float   measure_voltage(void);

int     uart_getc_nb(void);
int     uart_getc(void);
uint8_t uart_gets(char * buf, uint8_t size);

void    uart_putc(const char);
void    uart_puts(const char *);
#define uart_puts_P(s) (uart_puts((const char*)s));

void    read_calibration(void);
void    write_calibration(void);

void    do_calibrate(void);
void    do_discharge(void);
void    do_status(void);
void    do_help(void);


/* global variables */

volatile uint8_t tick; // seconds, counts down

float cal1; // Ux = ADC * cal1
float cal2; // Ix = Ux / cal2


/* constants */

const char progress[] = ".oOo._.oOo"; //progress indicator


/* constants in EEPROM */

EEMEM uint32_t ee_magic;
EEMEM float ee_cal1;
EEMEM float ee_cal2;


/* the code */

int __attribute__((OS_main)) main(void)
{

    hw_init();
    read_calibration();
    lcd_init(LCD_DISP_ON);

    lcd_clrscr();
    lcd_puts_P(
               "LiIon Tester 1.0\n"
               "XL 2016/03/02"
              );

    uart_puts_P("Li-Ion Akku Tester 1.0 - XL 2016/03/02\n");
    do_help();

    sei();

    while (1) {
        int c= uart_getc_nb();
        if (c >= 0) {
            switch ((char)c) {
            case 'c':
                do_calibrate();
                break;
            case 'd':
                do_discharge();
                break;
            case 's':
                do_status();
                break;
            default:
                uart_puts_P("\n?error\n");
                do_help();
            }
        }
        sleep_cpu();
    }
}


/* Interrupt Service Routines */

/* Timer 1 overflow - once per second */
ISR(TIMER1_COMPA_vect)
{
    uint8_t t= tick;
    if (t) {
        t--;
        tick= t;
    }
}


/* empty ISR for UART RX and ADC allows enabling interrupts
 * for the cpu to leave sleep mode on operation complete */

EMPTY_INTERRUPT(USART_RX_vect);
EMPTY_INTERRUPT(ADC_vect);



/* Sub Routines */

void hw_init(void)
{
    /* pullups enabled on all inputs */
    /* bus pins initialized as inputs */

    PORTD= 0xFF;
    DDRD=  0;

    PORTC= 0xFF;
    DDRC=  0;

    /* LED and DISCHARGE off */
    PORTB= ~(_BV(LCD_E_PIN) | _BV(LED_PIN) | _BV(DISCHARGE_PIN));
    DDRB=   (_BV(LCD_RS_PIN) | _BV(LCD_RW_PIN) | _BV(LCD_E_PIN) | _BV(LED_PIN) | _BV(DISCHARGE_PIN));

    /* disable Analog Comparator */
    ACSR = (_BV(ACD));

    /* ADC */
    /* internal Ref, input ADC7 */
    ADMUX = _BV(REFS1) | _BV(REFS0) | _BV(MUX2) | _BV(MUX1) | _BV(MUX0);
    /* disable ADC, enable IRQ, clk = clk_io/128 = 125kHz */
    ADCSRA = _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);

    /* USART0 */
    /* RX IRQ enable, RX enable, TX enable */
    UCSR0B = _BV(RXCIE0) | _BV(RXEN0) | _BV(TXEN0);
    /* word format 8N1 */
    UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);
    /* Baud rate */
    UBRR0 = (F_CPU/16/BAUD) - 1;

    /* Timer 1: fire once a second */
    OCR1A = (F_CPU/256) - 1;
    TCCR1A = 0; // WGM1:3 = 0010 = CTC mode
    TCCR1B = _BV(WGM12) | _BV(CS12);  // CS1:2  = 100 = clk_io/256
    /* enable compare match interrupt */
    TIMSK1 = _BV(OCIE1A);

    /* enable idle sleeping */
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
}


float measure_voltage(void)
{
    // enable ADC, start dummy conversion
    ADCSRA|= (_BV(ADEN) | _BV(ADSC));
    while (ADCSRA & _BV(ADSC))
        ;

    // average next 10 conversions
    uint16_t res= 0;
    for (uint8_t i=0; i<10; i++) {
        // start conversion by going idle
        sleep_cpu();
        while (ADCSRA & _BV(ADSC))
            ;
        res+= ADC;
    }

    //disable ADC
    ADCSRA&= ~_BV(ADEN);

    return cal1*res/10.0;
}


int uart_getc_nb(void)
{
    if (UCSR0A & _BV(RXC0))
        return uart_getc();
    return -1; // nothing to be read
}


int uart_getc(void)
{
    while (!(UCSR0A & _BV(RXC0)))
        ;
    uint8_t st= UCSR0A;
    uint8_t c= UDR0;
    if ((st & (_BV(FE0)|_BV(DOR0))) == 0) {
        UDR0= c; // echo character
        return (int)c;
    }
    return -2; // frame error or overrun
}


void uart_putc(const char c)
{
    // convert '\n' to "\r\n"
    if (c == '\n')
        uart_putc('\r');

    while ((UCSR0A & _BV(UDRE0)) == 0)
        ;

    UDR0= c;
}


void uart_puts(const char *s)
{
    while (*s) {
        uart_putc(*s);
        s++;
    }
}


uint8_t uart_gets(char * buf, uint8_t n)
{
    n--;
    buf[n]= '\0';
    uint8_t i=0;
    for ( ; i<n; i++) {
        int c= uart_getc();
        if (c<0 || (char)c == '\r') {
            buf[i]= '\0';
            return i;
        }
        buf[i]= (char)c;
    }
    return i;
}


void do_calibrate(void)
{
    char buf[8];
    float u1= measure_voltage();
    float u2;

    uart_puts_P("\ncalibration. U=");
    snprintf(buf, 8, "%.3f", u1);
    uart_puts(buf);
    uart_puts_P("V\nYour value (V): ");

    if (uart_gets(buf, 8) && sscanf(buf, "%f", &u2)) {
        cal1= cal1*u2/u1;
    } else {
        uart_puts_P("\ncould not parse '");
        uart_puts(buf);
        uart_puts_P("'\n> ");
        return;
    }

    u1= measure_voltage();
    uart_puts_P("\nrecalibrated. U=");
    snprintf(buf, 8, "%.3f", u1);
    uart_puts(buf);
    uart_puts_P("V\nLoad resistor (Ohm): ");

    if (uart_gets(buf, 8) && sscanf(buf, "%f", &u2)) {
        cal2= u2;
        snprintf(buf, 8, "%.2f", cal2);
        uart_puts_P("\nR=");
        uart_puts(buf);
        uart_puts_P("Ohm\n");
    } else {
        uart_puts_P("\ncould not parse '");
        uart_puts(buf);
        uart_puts_P("'\n> ");
        return;
    }

    write_calibration();
    uart_puts_P("calibration done\n> ");
}


void read_calibration(void)
{
    uint32_t magic= eeprom_read_dword(&ee_magic);
    if (magic == MAGIC) {
        cal1= eeprom_read_float(&ee_cal1);
        cal2= eeprom_read_float(&ee_cal2);
    } else {
        /* resistor divider 47K:15K, 1.1V int. Uref */
        cal1= 1.1*(47+15)/15/1023;
        /* default load resistor: 5 Ohm */
        cal2= 5.0;
    }
}


void write_calibration(void)
{
    eeprom_write_dword(&ee_magic, MAGIC);
    eeprom_write_float(&ee_cal1, cal1);
    eeprom_write_float(&ee_cal2, cal2);
}


void do_discharge(void)
{
    float u;
    char buf[8];

    uart_puts_P("\ndischarge requested\n");
    lcd_clrscr();

    u= measure_voltage();
    if (u <= LOW_VOLTAGE) {
        snprintf(buf, 8, "%.2f", u);

        uart_puts_P("voltage too low! U=");
        uart_puts(buf); uart_puts_P("V\n> ");

        lcd_puts_P("voltage too low!\nU=");
        lcd_puts(buf); lcd_putc('V');

        return;
    }

    tick= 1;
    while (tick) {
        sleep_cpu();
    }

    DISCHARGE_on();
    LED_on();
    uart_puts_P(
                "discharge started - stop with CTRL+C or 'q'\n\n"
                "seconds\tvolt\tampere\tAh\tWh\n"
               );
    lcd_puts_P("discharging");

    float Ah= 0;
    float Wh= 0;
    float i= 0;
    unsigned long t= 0;
    uint8_t last_tick= 0;

    while (1) {
        tick= 10;
        while (tick) {
            if (last_tick != tick) {
                lcd_gotoxy(15, 0);
                lcd_putc(progress[last_tick-1]);
                last_tick= tick;
            }
            int c= uart_getc_nb();
            if (c >= 0 && ((char)c == 'q' || (char)c == 3)) {
                DISCHARGE_off();
                LED_off();
                uart_puts_P(" - discharging stopped\n> ");
                return;
            }
            sleep_cpu();
        }

        u= measure_voltage();
        t+= 10;
        snprintf(buf, 8, "%lu", t);
        uart_puts(buf); uart_putc('\t');

        snprintf(buf, 8, "%.3f", u);
        uart_puts(buf); uart_putc('\t');
        lcd_clrscr(); lcd_puts(buf); lcd_puts_P("V ");

        i= u/cal2;
        snprintf(buf, 8, "%.3f", i);
        uart_puts(buf); uart_putc('\t');
        lcd_puts(buf); lcd_puts_P("A\n");

        Ah+= i/360;
        snprintf(buf, 8, "%.4f", Ah);
        uart_puts(buf); uart_putc('\t');
        snprintf(buf, 8, "%.3f", Ah);
        lcd_puts(buf); lcd_puts_P("Ah ");

        Wh+= u*i/360;
        snprintf(buf, 8, "%.4f", Wh);
        uart_puts(buf); uart_putc('\n');
        snprintf(buf, 8, "%.3f", Wh);
        lcd_puts(buf); lcd_puts_P("Wh");

        if (u < LOW_VOLTAGE) {
            DISCHARGE_off();
            LED_off();
            uart_puts_P("discharging ended\n> ");
            lcd_gotoxy(15, 0); lcd_putc('F');
            return;
        }
    }
}


void do_status(void)
{
    char buf[8];
    float u= measure_voltage();
    snprintf(buf, 8, "%.3f", u);

    lcd_clrscr();
    lcd_puts_P("U="); lcd_puts(buf); lcd_putc('V');
    uart_puts_P("\nU="); uart_puts(buf);

    snprintf(buf, 8, "%.2f", cal2);
    lcd_puts_P("\nR="); lcd_puts(buf); lcd_putc(OMEGA);
    uart_puts_P("V, R="); uart_puts(buf); uart_puts_P("Ohm\n> ");
}


void do_help(void)
{
    uart_puts_P(
                "Press 'c' to calibrate\n"
                "Press 'd' to start discharge\n"
                "Press 's' to show status\n> "
               );
}

